/**ElevenServer.java
 * A primitive database.  This will accept transactions through network
 * connections.  Security is very low.  Transactions will be
 * logged, and the IP will be logged as well, but no password is required,
 * and only light verification is performed that the user is in fact a user.

 * Students will write a program whose main purpose is similar to the original BankAccount.
 * Will take in a file, will receive requests to modify accounts in that file.  File I/O
 * will be written for them, as well as a command that will send to server.

P3--ElevenBank
    request balance, transfer
    userinfo set
P4--Buddy list set, retrieve, and get balances of.
    userinfo retrieve, usage
*/

import java.net.*;
import java.io.*;
import java.lang.Integer;
import java.util.Calendar;

public class ElevenServer {
  public static final int DEFAULT_PORT=27611; //which isn't taken, according to /net/services
  public static final int DEFAULT_BACKLOG=100;//This is sort of cruel, but it pretty much
  //guarantees no client gets rejected.
  public static final String OUTPUTLOGFILE = "DBLog.txt"; //name of log file--transactions go here
  public static final String ERRORLOGFILE = "ErrLog.txt"; //errors here.

  private static final String DELIM = ElevenDB.DELIM;
  private static final String MAGIC = "fureai kippu";

  private PrintWriter transLog;   //actual logs
  private PrintWriter errLog;     
  private ServerSocket listener;
  PrintWriter outputLog; //where to log all transactions.
  /*Data structures to hold the database.*/
  private ElevenDB theDB;  //use a hashtable if need to scale?

  /*Default constructor--empty DB*/
  public ElevenServer() {
    theDB = new ElevenDB();
    try {
      listener = new ServerSocket(DEFAULT_PORT, DEFAULT_BACKLOG);
      transLog = new PrintWriter(new BufferedWriter(new FileWriter(OUTPUTLOGFILE, true)));
      errLog = new PrintWriter(new BufferedWriter(new FileWriter(ERRORLOGFILE, true)));
    } catch (IOException e) {
      error(e.getMessage());
      error("Error on server construction!");
    }
  }
  /*Construct with DB file.  Read, parse, store into memory.*/
  public ElevenServer(String filename, boolean isText) {
    this();
    theDB = new ElevenDB(filename, isText);
  }

  /*Construct with DB and transaction log file.  Read, parse, and store
   *into memory.  If reading from a serialized output, then there's no
   *need to go through the logs--it's stored in the serialized format.*/
  public ElevenServer(String logfile, String filename, boolean isText) {
    this();
    theDB = new ElevenDB(filename, isText);
    try {
      BufferedReader buff = new BufferedReader(new FileReader (logfile));
      String input = buff.readLine();
      while (input != null) {
	String [] parsed = input.split(DELIM);
	if (parsed.length < 3) {input = buff.readLine(); continue;}
	System.out.println(parsed[2]);
	if (parsed[2].equals("TRANSFER")) {
	  System.out.println("Adding history");
	  theDB.addToHistory(parsed[1], parsed[1], parsed[3], 
			     Integer.parseInt(parsed[4]), parsed[5], 
			     parsed[0]); //date is at the end.
	  theDB.addToHistory(parsed[3], parsed[1], parsed[3], 
			     Integer.parseInt(parsed[4]), parsed[5], 
			     parsed[0]); //likewise.
	}
		
	input = buff.readLine();
      }
    } catch (IOException e) {
      error(e.getMessage());
      error("Error on server construction!");
    }


  }
  /*Starts the server.  Waits for incoming attempts to connect, and then 
   * handles their request.  Non-threaded, but transactions will be 
   * swift, to the point, and infrequent.  Make it threaded if this
   * goes beyond the scope of a class project.*/
  public void start() {
    //init messages
    transLog.println(Calendar.getInstance().getTime() + "\tServer started");
    transLog.flush();
    errLog.println(Calendar.getInstance().getTime() + "\tServer started");
    errLog.flush();
    while(true) {
      try {
	System.out.println("Waiting for connection");
	Socket comm = listener.accept();
	handleClient(comm);
	comm.close();
      } catch (IOException e) {
	error(e.getMessage());
	error("Ignoring error with client.  Listening for next connection.");
      }  catch (Exception e) {
	error(e.getMessage());
	e.printStackTrace();
	error("Unexpected error.  Saving DB to DBDump.txt, contuing.");
	theDB.writeOut("DBDump.txt", false);
      } catch (Error e) {
	error(e.getMessage());
	e.printStackTrace();
	error("Unexpected unrecoverable error.  Attempting to save DB.");
	theDB.writeOut("DBDump.txt", false);
	throw e;
      }
    }
  }
  /*Reads info from a socket about a transaction.  Information will be 
   * Read off of a BufferedReader stream, converted into a String.  The
   * String will have three pieces of data--login, transaction-type, and
   * amount of transaction.*/
  private void handleClient(Socket comm) throws IOException {
    String inputLine;
    String [] parsedData;
    BufferedReader input = new BufferedReader(new InputStreamReader(new DataInputStream(comm.getInputStream())));
    PrintWriter output = new PrintWriter(new BufferedWriter(new OutputStreamWriter(new DataOutputStream(comm.getOutputStream()))));
    inputLine = input.readLine();
    System.out.println(inputLine);//debug
    if (!checkArgs(output, inputLine)) {
      return;
    }

    parsedData = inputLine.split(DELIM, -1);  //tab delimeter, keep trailing empties

    String logString = "";
    String replyString = "";
    /**Dispatch method.  This is getting awfully long--time to reconstruct
     * and regeneralize, but too close to the wire for this.*/
    try {
      //-----------------------------TRANSFER-----------------------------------//

      if (parsedData[2].equals("TRANSFER")) {
	//magic user TRANSFER user_to_transfer_to amount comment
	if ((Integer.parseInt(parsedData[4]) < 0) || //prevent negative transfers
	    (modifyDB(parsedData[3], "BALANCE", "0") == -1) || //invalid second acct
	    (modifyDB(parsedData[1], "BALANCE", "0") < Integer.parseInt(parsedData[4]))) {//overdraw
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
	modifyDB(parsedData[1], "WITHDRAW", parsedData[4]);
	modifyDB(parsedData[3], "DEPOSIT", parsedData[4]);
	reply("TRANSACTION CONFIRMED", output);
	System.out.println("reply:TRANSACTION CONFIRMED"); //debug
	logString = parsedData[1] + DELIM + "TRANSFER" + DELIM + parsedData[3] +DELIM + parsedData[4] + DELIM + parsedData[5];
	//add transaction info to both parties
	theDB.addToHistory(parsedData[1], parsedData[1], parsedData[3], 
			   Integer.parseInt(parsedData[4]), parsedData[5]);
	theDB.addToHistory(parsedData[3], parsedData[1], parsedData[3], 
			   Integer.parseInt(parsedData[4]), parsedData[5]);

      } else if ((parsedData[2].equals("WITHDRAW")) ||
		 (parsedData[2].equals("DEPOSIT"))) {
	//these actions disabled to outside.
	reply("ERROR", output);
	System.out.println("reply:ERROR"); //debug
	return;

	//-----------------------------BALANCE-----------------------------------//

      } else if (parsedData[2].equals("BALANCE")) {
	//magic user BALANCE
	if (parsedData.length == 3) {
                  
	  int balance = modifyDB(parsedData[1], "BALANCE", "0");
	  if (balance == -1) {
	    reply("ERROR", output);
	    System.out.println("reply:ERROR"); //debug
	    return;
	  }
	  reply("BALANCE" + DELIM + balance, output);
	  System.out.println("reply:BALANCE" + DELIM + balance); //debug
	  logString = parsedData[1] + DELIM  + "BALANCE";
	} else {
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}

	//-----------------------GETINFO AND GETDETAILS----------------------------//

      } else if ((parsedData[2].equals("GETINFO")) ||
		 (parsedData[2].equals("GETDETAILS"))){
	//get information about user from the DB
	boolean namePass = parsedData[2].equals("GETINFO");
	String data;
	String printout = parsedData[2] + DELIM;  //GETINFO or GETDETAILS in the
	logString = parsedData[1] + DELIM+ parsedData[2] + DELIM; //strings here.
	switch (parsedData.length) {

	case 3:		             //magic user GETINFO
	  data = theDB.getInfo(parsedData[1], true, namePass);
	  printout += parsedData[1];
	  logString += parsedData[1];
	  break;

	case 4:		             //magic user GETINFO username
	  data = theDB.getInfo(parsedData[3], false, namePass);
	  printout += parsedData[3];
	  logString += parsedData[3];
	  break;

	case 5:                      //magic user GETINFO username password
	  String nameAndPassword = theDB.getInfo(parsedData[3], true, true);
	  data = theDB.getInfo(parsedData[3], true, namePass);
	  printout += parsedData[3] + DELIM + parsedData[4];
	  logString += parsedData[3] + DELIM + parsedData[4]; 
	  //password check
	  if (nameAndPassword.split(DELIM)[1].equals(parsedData[4])) 
	    break;

	default:
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
	if (data.equals("")) {
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
	System.out.println(printout);
	reply("USERINFO" + DELIM + data, output);

	//-----------------------------UPDATEINFO--------------------------------//

      } else if (parsedData[2].equals("UPDATEINFO")) {
	//magic user UPDATEINFO nick pass
	if (parsedData.length == 5) {
	  modifyDBInfo(parsedData[1], parsedData[3], parsedData[4]);
	  reply("UPDATE CONFIRMED", output);
	  System.out.println("reply:UPDATE CONFIRMED");
	  logString = parsedData[1] + DELIM + "UPDATEINFO" + DELIM + 
	    parsedData[3] + DELIM + parsedData[4];
	} else 
	  //magic user UPDATEINFO nick pass quote trades email screen
	  if (parsedData.length == 9) {
	    modifyDBInfo(parsedData[1], parsedData[3], parsedData[4]);
	    modifyDBDetail(parsedData[1], parsedData[5], parsedData[6],
			   parsedData[7], parsedData[8]);
	    reply("UPDATE CONFIRMED", output);
	    System.out.println("reply:UPDATE CONFIRMED");
	    logString = parsedData[1] + DELIM + "UPDATEINFO" + DELIM + parsedData[3] + DELIM
	      + parsedData[4] + DELIM + parsedData[5] + DELIM + parsedData[6]
	      + parsedData[7] + DELIM + parsedData[8];
	  } else {
	    reply("ERROR", output);
	    System.out.println("reply:ERROR"); //debug
	    return;
	  }

	//---------------------------GET_HISTORY---------------------------------//

      } else if (parsedData[2].equals("GET_HISTORY")) {
	//magic user HISTORY_GET
	ElevenDBTransaction trans = null;
	if ((theDB.getValue(parsedData[1])) != -1) {  //check if user exists
	  String transList = theDB.getHistory(parsedData[1]);
	  reply(transList, output);
	  System.out.println("reply:User's history");
	  logString = parsedData[1] + DELIM + "GET_HISTORY";
	} else {  //user doesn't exist
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
	/**********************************IMACCOUNT'S FUNCTIONS******************************/
	//----------------------------BUDDYLIST---------------------------------//
      } else if (parsedData[2].equals("BUDDYLIST")) {
	// magic user BUDDYLIST name pass
	String data = theDB.getInfo(parsedData[3], true, true);
	if ((parsedData.length == 5) &&
	    ((theDB.getValue(parsedData[1])) != -1) &&  //check both users exist
	    ((theDB.getValue(parsedData[3])) != -1) &&
	    (data.split(DELIM)[1].equals(parsedData[4]))) //password check
	  {
	    theDB.addToBuddyList(parsedData[1], parsedData[3]);

	    reply("BUDDYLIST UPDATED", output);
	    System.out.println("reply:BUDDYLIST UPDATED");
	    logString = parsedData[1] + DELIM + " BUDDYLIST" + DELIM + parsedData[3] + DELIM
	      + parsedData[4];
	  } else {
	    reply("ERROR", output);
	    System.out.println("reply:ERROR"); //debug
	    return;                
	  }
            
	//---------------------------LEAVEMESSAGE--------------------------------//
      } else if (parsedData[2].equals("LEAVEMESSAGE")) {
	// magic user LEAVEMESSAGE target message
	if ((parsedData.length == 5) &&
	    ((theDB.getValue(parsedData[1])) != -1) &&  //check both users exist
	    ((theDB.getValue(parsedData[3])) != -1) &&
	    (theDB.searchBuddies(parsedData[1], parsedData[3]))) {//buddylist check
	  theDB.addMessage(parsedData[3], parsedData[1], parsedData[4]);
	  reply("MESSAGE SENT", output);
	  System.out.println("reply:MESSAGE SENT");
	  logString = parsedData[1] + DELIM + "LEAVEMESSAGE" + DELIM + parsedData[3];
	} else {
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
	//---------------------------CHECKIMINFO---------------------------------//
      } else if (parsedData[2].equals("CHECKIMINFO")) {
	// magic user CHECKIMINFO
	String buddyInfo;
	String messages;
	if (parsedData.length == 3) {
	  buddyInfo = theDB.getBuddyInfos(parsedData[1]);
	  messages = theDB.getMessages(parsedData[1]);
	  reply("Buddy activity:\0" + buddyInfo + "\0Messsages:\0" +
		messages, output);
	  System.out.println("reply:BuddyInfo and Messages");
	  logString = parsedData[1] + DELIM + "CHECKIMINFO";
	} else {  //user doesn't exist
	  reply("ERROR", output);
	  System.out.println("reply:ERROR"); //debug
	  return;
	}
              
	//---------------------------WRITEOUT-----------------------------------//

      } else if (parsedData[2].equals("WRITEOUT")) {
	//magic user WRITEOUT
	//admin sort of command to save the DB.
	theDB.writeOut("savedDB.txt", true);
	theDB.writeOut("savedDB.serialized", false);
	reply("WRITING DB", output);
	System.out.println("WRITING DB");
	logString = parsedData[1] + DELIM + "WRITINGDB";	    
      } else {
	//error stuff
	reply("ERROR", output);
	System.out.println("reply:ERROR"); //debug
	return;
      }
    } catch (Exception e) {
      //Basically catches all I/O exceptions with client
      //replace with homemade exception
      System.out.println(e.getMessage());
      e.printStackTrace();
      reply("ERROR", output);
      System.out.println("reply:ERROR[exception]"); //debug
      return;	    
    }
    //On success, add to the transLog, update the lastLogin.
    theDB.updateLastLogin(parsedData[1]);
    transLog.println(Calendar.getInstance().getTime() + DELIM + logString +DELIM + comm.getInetAddress());
    transLog.flush();
  }


  /**Checks the arguments.  Prevents obvious errors from being used to modify the database.*/
  public boolean checkArgs(PrintWriter output, String inputLine) {
    if (inputLine == null) {
      System.err.println("Invalid message format.  Ignoring.");
      output.println("ERROR");
      output.flush();
      return false;
    }
    String [] parsedData = inputLine.split(DELIM);  //tab delimeter

    //check args
    if (parsedData.length < 3){
      System.err.println("Invalid number of arguments.  Ignoring.");
      output.println("ERROR");
      output.flush();
      return false;
    }
    /*check for primitive security*/
    if (!parsedData[0].equals(MAGIC)) {
      System.err.println("Invalid magic string.  Ignoring.");
      output.println("ERROR");
      output.flush();
      return false;
    }
    return true;
  }

  /**Given a user, transactionType, and amount, modify the internal data structure
   * to represent the appropriate changes.  Usernames must match exactly.
   * valid transactionTypes are:
   *    WITHDRAW - Decrease db entry by amount
   *    DEPOSIT  - Increase db entry by amount
   *    BALANCE  - get balance of this user.  If user does not exist
   *               return -1
   * amount must parse into an integer, which represents number of cents to 
   * deposit or withdraw.  It's ignored if the transaction is BALANCE.*/
  private int modifyDB(String username, String transactionType, String amount) {
    if (transactionType.equals("BALANCE")) {
      return theDB.getValue(username);
    } else if (transactionType.equals("WITHDRAW")) {
      theDB.modifyValue(username, -1 * Integer.parseInt(amount));
    } else if (transactionType.equals("DEPOSIT")) {
      theDB.modifyValue(username, Integer.parseInt(amount));
    } else {
      //throw an InvalidTransactionException
      return -30000;
    }
    return 0; //all good.
  }

  /** Get this user's info.  */

  /**Modify the user's user and pass.  Returns true if successful, false if the user
   * diesn't exist or some other error.*/
  private boolean modifyDBInfo(String username, String nick, String pwd) {
    
    return theDB.setInfo(username, nick, pwd);
  }

  /** Modify this user's details.  Returns true if successful, false if the user
   *  doesn't exist or some other mishap.*/
  private boolean modifyDBDetail(String username, String quote, String trades,
                                 String email, String screen) {
    return theDB.setDetails(username, quote, trades, email, screen);
  }
    
  public static void main(String args[]) {
    ElevenServer classServer;
    final boolean ISTEXT = true;
    if (args.length == 2) {//DB, logfile
      classServer = new ElevenServer(args[1], args[0], ISTEXT);
    } else if (args.length == 1) {
      classServer = new ElevenServer(args[0], ISTEXT);
    } else {
      classServer = new ElevenServer();
    }
    classServer.start();
  }


  /*Util functions*/

  /** Write an error to both System.err and errLog.  Throws the date
   *  in brackets on there too.*/
  private void error(String errStr) {
    System.err.println(Calendar.getInstance().getTime() + errStr);
    errLog.println(Calendar.getInstance().getTime() + errStr);
    errLog.flush();
  }

  /** Write a msg to the output stream.  Here it reduces two statements to one
   *  when replying to sockets.*/
  private void reply(String msg, PrintWriter out) {
    out.println(msg);
    out.flush();
  }
}
